package org.baxter.disco.ocr;

import java.io.File;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

/**
 * A single cycle in a test.
 *
 * @author Blizzard Finnegan
 * @version 1.0.0, 20 Mar. 2023
 */
public class Cycle
{
    /**
     * Current cycle number.
     */
    private final int cycleNumber;

    /**
     * Map of values, paired to their camera.
     *
     * Key: String of the name of the camera 
     * Value: Pair[File,Double] of the information for this cycle.
     *
     * Pair[File,Double]:
     * First: File of the location of the image from the camera.
     * Second: Double generated by Tesseract parsing the above image.
     */
    private final Map<String,Pair<File,Double>> cycleValues = new TreeMap<>();

    /**
     * Default constructor.
     */
    public Cycle(int cycleNumber)
    { this.cycleNumber = cycleNumber; }

    /**
     * Getter for this object's cycle number.
     *
     * @return int of the current cycle number.
     */
    public int getcycleNumber() { return this.cycleNumber; }

    /**
     * Get all cameras contained within this object.
     *
     * @return Set of Strings of all cameras available in this object.
     */
    public Set<String> getCameras() { return this.cycleValues.keySet(); }

    /**
     * Getter for the given camera's associated image File.
     *
     * @param cameraName    name of the camera to get the image from 
     *
     * @return File of the image from the camera
     */
    public File getImage(String cameraName) 
    { return this.cycleValues.get(cameraName).first(); }

    /**
     * Get the Tesseract parsed value of the image from the given camera.
     *
     * @param cameraName    name of the camera to get the image from 
     *
     * @return Double of the Tesseract-parsed value of the image.
     */
    public Double getValue(String cameraName) 
    { return this.cycleValues.get(cameraName).second(); }

    /**
     * Add a camera to the object, with its accociated image and value.
     *
     * @param cameraName    Name of the camera associated with the image and value
     * @param image         File object denoting the location of the image from the camera
     * @param parsedValue   Double of the value generated by Tesseract from the image
     */
    public void addCamera(String cameraName, File image, Double parsedValue)
    { this.cycleValues.put(cameraName,new Pair<File,Double>(image, parsedValue)); }
}
