package org.baxter.disco.ocr;

//Pi4J imports
import com.pi4j.Pi4J;
import com.pi4j.context.Context;
import com.pi4j.io.gpio.digital.DigitalInput;
import com.pi4j.io.gpio.digital.DigitalInputConfigBuilder;
import com.pi4j.io.gpio.digital.DigitalOutput;
import com.pi4j.io.gpio.digital.DigitalOutputConfigBuilder;
import com.pi4j.io.gpio.digital.DigitalState;
import com.pi4j.io.gpio.digital.PullResistance;

/**
 * Facade for all movement of the fixture.
 *
 * Uses Pi4J to communicate with GPIO pins.
 * Currently missing Run switch compatibility.
 *
 * @author Blizzard Finnegan
 * @version 3.3.0, 20 Mar. 2023
 */
public class MovementFacade
{
    /**
     * Boolean used to communicate with runSwitchThread to gracefully exit.
     */
    private static boolean exit = false;

    /**
     * Thread that watches the physical Run switch on the device so that 
     * fixture movement can be stopped.
     */
    private static Thread runSwitchThread;

    /**
     * Lock object for multithreading/run switch.
     */
    private static final Object lockObject = new Object();

    /**
     * Fraction of the total travel time, so the arm won't push through the limit switch.
     */
    private static final double SLOW_POLL_FACTOR = 0.95;

    /**
     * Amount of distance to travel.
     * Measured in polls (~10ms intervals currently)
     * Set in {@link #findDistance()}
     */
    private static double TRAVEL_DIST;

    /**
     * How many milliseconds to wait before polling the GPIO
     */
    private static final int POLL_WAIT = 10;

    /**
     * How many polls before assuming bad GPIO connection.
     * The DUTs fall asleep after ~6 seconds,so the travel must be less than 3 seconds.
     */
    private static final int TIMEOUT = 300;


    //PWM Addresses
    //All addresses are in BCM format.

    /**
     * Output pin address for motor power control.
     */
    private static final int MOTOR_ENABLE_ADDR = 22;

    /**
     * Output pin address for motor direction control.
     */
    private static final int MOTOR_DIRECTION_ADDR = 27;

    /**
     * Output pin address for piston control.
     */
    private static final int PISTON_ADDR = 25;

    /**
     * Input pin address for the run switch. 
     */
    private static final int RUN_SWITCH_ADDR = 10;

    /**
     * Input pin address for the upper limit switch.
     */
    private static final int UPPER_LIMIT_ADDR = 23;

    /**
     * Input pin address for the lower limit switch.
     */
    private static final int LOWER_LIMIT_ADDR = 24;

    //Pi GPIO pin objects
    
    /**
     * <pre>
     * Upper limit switch object.
     * 
     * Status: High; Upper limit switch has been reached.
     * Status: Low; Upper limit switch has not been reached.
     * </pre>
     */
    private static DigitalInput upperLimit;

    /**
     * <pre>
     * Lower limit switch object.
     * 
     * Status: High; Lower limit switch has been reached.
     * Status: Low; Lower limit switch has not been reached.
     * </pre>
     */
    private static DigitalInput lowerLimit;

    /**
     * <pre>
     * Run switch object.
     *
     * Status: High; Test may continue.
     * Status: Low; Test must stop immediately.
     * </pre>
     */
    private static DigitalInput runSwitch;

    /**
     * <pre>
     * Motor power object.
     *
     * Status: High; Motor starts moving, in the direction defined by {@link #motorDirection}.
     * Status: Low; Motor stops moving.
     * </pre>
     */
    private static DigitalOutput motorEnable;

    /**
     * <pre>
     * Defines the movement direction for the motor enabled by {@link #motorEnable}.
     *
     * Status: High; Motor will move upwards.
     * Status: Low; Motor will move downwards.
     * </pre>
     */
    private static DigitalOutput motorDirection;

    /**
     * <pre>
     * Piston control pin object.
     *
     * Status: High; Piston is extended.
     * Status: Low; Piston is retracted.
     * </pre>
     */
    private static DigitalOutput pistonActivate;

    /**
     * {@link Pi4J} API interaction object.
     */
    private static Context pi4j;

    public static void init() throws Exception
    {
        pi4j = Pi4J.newAutoContext();

        ErrorLogging.logError("DEBUG: Opening input GPIO pins...");
        upperLimit = inputBuilder(UPPER_LIMIT_ADDR);
        lowerLimit = inputBuilder(LOWER_LIMIT_ADDR);
        runSwitch  = inputBuilder(RUN_SWITCH_ADDR);

        ErrorLogging.logError("DEBUG: Opening output GPIO pins...");
        motorEnable    = outputBuilder(MOTOR_ENABLE_ADDR);
        motorDirection = outputBuilder(MOTOR_DIRECTION_ADDR);
        pistonActivate = outputBuilder(PISTON_ADDR);

        try
        {
            findDistance();
            pressButton();
        }
        catch(Exception e)
        {
            ErrorLogging.logError(e);
            pi4j.shutdown();
            throw new Exception("GPIO init error!!! Check GPIO connections.");
        }

        ErrorLogging.logError("DEBUG: Starting lock thread...");
        runSwitchThread = new Thread(() -> 
                {
                    while(!exit)
                    {
                        if(runSwitch.isOn())
                        {
                            synchronized(lockObject)
                            {
                                while(runSwitch.isOn())
                                {
                                    try{ Thread.sleep(100); } catch(Exception e){ErrorLogging.logError(e);}
                                    if(runSwitch.isOff())
                                    {
                                        try{ Thread.sleep(100); } catch(Exception e){ErrorLogging.logError(e);}
                                        if(runSwitch.isOff())
                                        {
                                            break;
                                        }
                                    }
                                }
                            }
                        }
                        try{ Thread.sleep(1); } catch(Exception e){ErrorLogging.logError(e);}
                    }
                }, "Run switch monitor.");
        runSwitchThread.start();

    }


    /**
     * Builder function for DigitalInput pins. 
     *
     * @param address   BCM address of the {@link DigitalInput} pin.
     *
     * @return newly created {@link DigitalInput} object.
     */
    private static DigitalInput inputBuilder(int address)
    { 
        DigitalInputConfigBuilder configBuilder = DigitalInput.newConfigBuilder(pi4j)
                                                              .address(address)
                                                              .pull(PullResistance.PULL_DOWN)
                                                              .debounce(3000L)
                                                              .provider("pigpio-digital-input");
        return pi4j.create(configBuilder);
    }

    /**
     * Builder function for DigitalOutput pins. 
     *
     * @param address   BCM address of the {@link DigitalOutput} pin.
     *
     * @return newly created {@link DigitalOutput} object
     */
    private static DigitalOutput outputBuilder(int address)
    {
        DigitalOutputConfigBuilder configBuilder = DigitalOutput.newConfigBuilder(pi4j)
                                                                .address(address)
                                                                .shutdown(DigitalState.LOW)
                                                                .initial(DigitalState.LOW)
                                                                .provider("pigpio-digital-output");
        return pi4j.create(configBuilder);

    }

    /**
     * Function used to locate the fixture's motor.
     */
    public static int resetArm()
    {
        ErrorLogging.logError("DEBUG: --------------------------------------");
        ErrorLogging.logError("DEBUG: Resetting arm...");
        int counter;
        if(upperLimit.isHigh())
        {
            ErrorLogging.logError("DEBUG: Motor at highest point! Lowering to reset.");
            synchronized(lockObject)
            {
                motorDirectionDown();
                motorOn();
                try{ Thread.sleep(500); }
                catch (Exception e){ ErrorLogging.logError(e); }
                motorOff();
            }
        }
        ErrorLogging.logError("DEBUG: Moving motor to highest point.");
        motorDirectionUp();

        synchronized(lockObject)
        {
            motorOn();

            for(counter = 0; counter < TIMEOUT; counter++)
            { 
                try{ Thread.sleep(POLL_WAIT); } catch(Exception e){ ErrorLogging.logError(e); }
                if(upperLimit.isOn()) 
                {
                    try{ Thread.sleep(1); } catch(Exception e){ErrorLogging.logError(e); }
                    if(upperLimit.isOn()) break;
                }
            }
            motorOff();
        }

        if(counter < TIMEOUT)
        {
            ErrorLogging.logError("DEBUG: Motor returned after " + counter + " polls.");
            ErrorLogging.logError("DEBUG: --------------------------------------");
            return counter;
        }

        else
        {
            ErrorLogging.logError("DEBUG: No motor return after 3 seconds.");
            ErrorLogging.logError("DEBUG: --------------------------------------");
            return -1;
        }
    }

    /**
     * Used to programmatically find the distance between the upper and lower limit switches.
     */
    private static void findDistance() throws Exception
    {
        ErrorLogging.logError("DEBUG: Measuring travel time to buttons...");
        int safeGPIO = resetArm();
        if(safeGPIO < 0) throw new Exception("Failed GPIO initialisation!");
        int downTravelCounter = 0;
        int upTravelCounter = 0;
        motorDirectionDown();
        motorOn();
        for(downTravelCounter = 0; downTravelCounter < TIMEOUT; downTravelCounter++)
        {
            try{ Thread.sleep(POLL_WAIT); } catch(Exception e){ ErrorLogging.logError(e); }
            if(lowerLimit.isOn()) 
            {
                try{ Thread.sleep(1); } catch(Exception e){ErrorLogging.logError(e); }
                if(lowerLimit.isOn()) break;
            }
        }
        motorOff();
        if(lowerLimit.isOff()) ErrorLogging.logError("DEBUG: False positive on findDistance down!");
        if(downTravelCounter == TIMEOUT) throw new Exception("PWM set too slow!");
        
        ErrorLogging.logError("DEBUG: Down travel count: " + downTravelCounter);

        motorDirectionUp();
        motorOn();
        for(upTravelCounter = 0; upTravelCounter < TIMEOUT; upTravelCounter++)
        { 
            try{ Thread.sleep(POLL_WAIT); } catch(Exception e){ ErrorLogging.logError(e); }
            if(upperLimit.isOn()) 
            {
                try{ Thread.sleep(1); } catch(Exception e){ErrorLogging.logError(e); }
                if(upperLimit.isOn()) break;
            }
        }
        motorOff();
        if(upperLimit.isOff()) ErrorLogging.logError("DEBUG: False positive on findDistance up!");
        if(upTravelCounter == TIMEOUT) throw new Exception("Failed GPIO initialisation!");

        ErrorLogging.logError("DEBUG: Up travel count: " + downTravelCounter);

        int travelCounter = Math.min(upTravelCounter, downTravelCounter);
        TRAVEL_DIST = travelCounter;
    }

    /**
     * Internal function to send the fixture to a given limit switch.
     *
     * Detects if the limit switch is active before activating motor.
     *
     * @param moveUp    Whether to send the fixture up or down. (True = up, False = down)
     * @return {@link FinalState} of the final status of the motor.
     * FAILED if travel fails. (Theoretically possible, no return state though)
     * UNSAFE if travel is stopped by the limit switch.
     * SAFE if travel is stopped by timeout, or if fixture is already at limit switch before movement.
     */
    private static FinalState gotoLimit(boolean moveUp)
    {
        FinalState output = FinalState.FAILED;
        DigitalInput limitSense;
        if(moveUp)  
        {
            motorDirectionUp();
            limitSense = upperLimit;
        }
        else        
        {
            motorDirectionDown();
            limitSense = lowerLimit;
        }

        if(limitSense.isOn()) return FinalState.SAFE;

        int totalPollCount = (int)(TRAVEL_DIST);
        int highSpeedPolls = (int)(totalPollCount * SLOW_POLL_FACTOR);
        ErrorLogging.logError("DEBUG: =============================");
        ErrorLogging.logError("DEBUG: Travel time: " + totalPollCount);
        ErrorLogging.logError("DEBUG: High speed poll count: " + highSpeedPolls);
        ErrorLogging.logError("DEBUG: =============================");

        synchronized(lockObject)
        {
            motorOn();
            for(int i = 0; i < highSpeedPolls; i++)
            {
                try{ Thread.sleep(POLL_WAIT); } catch(Exception e){ ErrorLogging.logError(e); }
                if(limitSense.isOn())
                {
                    try{ Thread.sleep(5); } catch(Exception e){ErrorLogging.logError(e); }
                    if(limitSense.isOn()) 
                    {
                        motorOff();
                        break;
                    }
                }
            }
            motorOff();
        }

        output = (limitSense.isOn() ? FinalState.UNSAFE : FinalState.SAFE);

        return output;
    }

    /**
     * Send the fixture to the lower limit switch.
     *
     * @return true if movement was successful; otherwise false
     */
    public static FinalState goDown() { return gotoLimit(false); }

    /**
     * Send the fixture to the upper limit switch.
     *
     * @return true if movement was successful; otherwise false
     */
    public static FinalState goUp() { return gotoLimit(true); }

    /**
     * Extends the piston for 1 second, pushing the button on the DUT.
     */
    public static void pressButton()
    {
        synchronized(lockObject)
        {
            ErrorLogging.logError("DEBUG: Pressing button...");
            pistonActivate.on();
            try{ Thread.sleep(1000); } catch(Exception e) {ErrorLogging.logError(e);};
            ErrorLogging.logError("DEBUG: Releasing button...");
            pistonActivate.off();
        }
    }

    /**
     * Closes connections to all GPIO pins.
     */
    public static void closeGPIO()
    {
        resetArm();
        if(runSwitchThread.isAlive())
        {
                exit = true;
                try{ Thread.sleep(500); } catch(Exception e){}
        }
        pi4j.shutdown();
    }

    /**
     * Function to move the fixture once for an iteration.
     *
     * @param prime     Whether or not to wake up the DUT
     */
    public static void iterationMovement(boolean prime)
    {
        goUp();
        if(prime) pressButton();
        goDown();
        try{ Thread.sleep(100); } catch(Exception e){ ErrorLogging.logError(e); }
        pressButton();
    }

    private static void motorOn()
    {
        ErrorLogging.logError("DEBUG: Motor on.");
        motorEnable.on();
    }

    private static void motorOff()
    {
        ErrorLogging.logError("DEBUG: Motor off.");
        motorEnable.off();
    }

    private static void motorDirectionUp()
    {
        ErrorLogging.logError("DEBUG: Motor travelling up.");
        motorDirection.high();
    }

    private static void motorDirectionDown()
    {
        ErrorLogging.logError("DEBUG: Motor travelling down.");
        motorDirection.low();
    }

    /**
     * Possible states for motor control movement.
     */
    public enum FinalState
    { UNSAFE, SAFE, FAILED; }
}
